package desmoj.tutorial2.InterruptsExample;

import desmoj.core.simulator.*;
import desmoj.core.dist.*;
import java.util.concurrent.TimeUnit;

/**
 * This is the model class. It is the main class of a process-oriented model of
 * the loading zone of a container terminal where urgent trucks are always
 * serviced first . Trucks arrive at the terminal to load containers. They
 * usually wait in line until a van carrier is available to fetch a certain
 * container and load it onto the truck. If the truck is urgent (has a high
 * priority) it doesn't have to wait but interrupts the service of a
 * non-priority truck. After loading is completed, the truck leaves the terminal
 * while the van carrier serves the next truck.
 * 
 * @author Olaf Neidhardt, Ruth Meyer
 */
public class InterruptsExample extends Model {

    /**
     * model parameter: the number of van carriers
     */
    protected static int NUM_VC = 1;

    /**
     * Random number stream used to draw an arrival time for the next truck. See
     * init() method for stream parameters.
     */
    private ContDistExponential truckArrivalTime;

    /**
     * Random number stream used to draw a service time for a truck. Describes
     * the time needed by the VC to fetch and load the container onto the truck.
     * See init() method for stream parameters.
     */
    private ContDistUniform serviceTime;

    /**
     * Random number stream used to determine if a truck is urgent. See init()
     * method for stream parameters.
     */
    private BoolDistBernoulli urgentStream;

    /**
     * A waiting queue object is used to represent the parking area for the
     * trucks. Every time a truck arrives it is inserted into this queue (it
     * parks) and will be removed by the VC for service.
     * 
     * This way all necessary basic statistics are monitored by the queue.
     */
    protected ProcessQueue<Truck> truckQueue;

    /**
     * A waiting queue object is used to represent the parking spot for the VC.
     * If there is no truck waiting for service the VC will return here and wait
     * for the next truck to come.
     * 
     * This way all idle time statistics of the VC are monitored by the queue.
     */
    protected ProcessQueue<VanCarrier> idleVCQueue;

    /**
     * A waiting queue object is used to store the busy VCs. This is needed in
     * order for an urgent truck to be able to interrupt a VC if there is no
     * idle VC available.
     */
    protected ProcessQueue<VanCarrier> busyVCQueue;

    /**
     * An interrupt code to tell the interrupted VC that an urgent truck is
     * awaiting service.
     */
    protected InterruptCode urgentTruckArrived;

    /**
     * InterruptsExample constructor.
     * 
     * Creates a new InterruptsExample model via calling the constructor of the
     * superclass.
     * 
     * @param owner
     *            the model this model is part of (set to <tt>null</tt> when
     *            there is no such model)
     * @param modelName
     *            this model's name
     * @param showInReport
     *            flag to indicate if this model shall produce output to the
     *            report file
     * @param showInTrace
     *            flag to indicate if this model shall produce output to the
     *            trace file
     */
    public InterruptsExample(Model owner, String modelName, boolean showInReport,
            boolean showInTrace) {
        super(owner, modelName, showInReport, showInTrace);
    }

    /**
     * returns a description of the model to be used in the report.
     * 
     * @return model description as a string
     */
    public String description() {
        return "This model describes a queueing system located at a "
                + "container terminal. Trucks will arrive and "
                + "require the loading of a container. A van carrier (VC) is "
                + "on duty and will head off to find the required container "
                + "in the storage. It will then load the container onto the "
                + "truck. Afterwards, the truck leaves the terminal. "
                + "In case the VC is busy, the truck waits " + "for its turn on the parking-lot. "
                + "If the VC is idle, it waits on its own parking spot for the " + "truck to come.";
    }

    /**
     * activates dynamic model components (simulation processes).
     * 
     * This method is used to place all events or processes on the internal
     * event list of the simulator which are necessary to start the simulation.
     * 
     * In this case, the truck generator and the van carrier(s) have to be
     * created and activated.
     */
    public void doInitialSchedules() {

        // create and activate the van carrier(s)
        for (int i = 0; i < NUM_VC; i++) {
            new VanCarrier(this, "VanCarrier", true).activate();
        }

        // create and activate the truck generator process
        new TruckGenerator(this, "TruckArrival", false).activate();
    }

    /**
     * initialises static model components like distributions and queues.
     */
    public void init() {

        // initalising the serviceTimeStream
        serviceTime = new ContDistUniform(this, "ServiceTimeStream", 3.0, 7.0, true, false);

        // initalising the truckArrivalTimeStream
        truckArrivalTime = new ContDistExponential(this, "TruckArrivalTimeStream", 3.0, true, false);

        // initalising the urgentStream
        urgentStream = new BoolDistBernoulli(this, "UrgentTruckStream", 0.1, true, false);

        // initalising the truckQueue
        truckQueue = new ProcessQueue<Truck>(this, "Truck Queue", true, true);

        // initalising the idleVCQueue
        idleVCQueue = new ProcessQueue<VanCarrier>(this, "idle VC Queue", true, true);

        // initalising the busyVCQueue
        busyVCQueue = new ProcessQueue<VanCarrier>(this, "busy VC Queue", true, true);

        // initialises the interrupt code
        urgentTruckArrived = new InterruptCode("Urgent Truck Arrived");
    }

    /**
     * Returns a sample of the random stream used to determine the time needed
     * to fetch the container for a truck from the storage area and the time the
     * VC needs to load it onto the truck.
     * 
     * @return double a serviceTime sample
     */
    public double getServiceTime() {
        return serviceTime.sample();
    }

    /**
     * Returns a sample of the random stream used to determine the next truck
     * arrival time.
     * 
     * @return double a truckArrivalTime sample
     */
    public double getTruckArrivalTime() {
        return truckArrivalTime.sample();
    }

    /**
     * Returns a sample of the random stream used to determine if a truck is
     * urgent.
     * 
     * @return boolean <tt>true</tt> if the truck is urgent; <tt>false</tt>
     *         otherwise
     */
    public boolean getTruckIsUrgent() {
        return urgentStream.sample();
    }

    /** run the model */
    public static void main(java.lang.String[] args) {

        // create model and experiment
        InterruptsExample model = new InterruptsExample(null, "InterruptsExample", true, true);
        Experiment exp = new Experiment("InterruptsExampleExperiment", TimeUnit.SECONDS,
                TimeUnit.MINUTES, null);

        // connect both
        model.connectToExperiment(exp);

        // set experiment parameters
        exp.setShowProgressBar(true); // display a progress bar (or not)
        exp.stop(new TimeInstant(1500)); // set end of simulation at 1500
                                         // minutes
        exp.tracePeriod(new TimeInstant(0), new TimeInstant(100)); // set the
                                                                   // period of
                                                                   // the trace
        exp.debugPeriod(new TimeInstant(0), new TimeInstant(50)); // and debug
                                                                  // output

        // start the Experiment at simulation time 0.0
        exp.start();

        // --> now the simulation is running until it reaches its end criterium
        // ...
        // ...
        // <-- afterwards, the main thread returns here

        // generate the report (and other output files)
        exp.report();

        // stop all threads still alive and close all output files
        exp.finish();
    }
} /* end of model class */